=begin
(c) TIG 2009
extrudeEdgesByLathe.rb
First Select a 'polyline', i.e. an arc of welded curves/arcs/lines etc.
OR a Face [then the Face's Edges will be used]
Choose the tool 'Extrude Edges by Lathe' on the Plugins menu 
or type 'extrudeEdgesByLathe' in the Ruby Console
or use the 'Extrude Edges By Lathe' on the 'Extrusion Tools' Toolbar.
A dialog asks for the arc's swept-angle in degrees (-ve = clockwise) 
and the number of the arc's segments, the initial defaults are 45.0 & 9.
You can choose to soft/smooth-edges (Yes|No|Yes+Smooth) default='No'.
Any changes are remembered for that model, across sessions.
You are then prompted to pick the arc's center-point.  
The VCB reports the dynamic point -e.g. [1.234,56.789,0.0].
You must then pick a second-point to set the axis-of-rotation - 
for example, picking a second-point vertically above the center-point 
(using inference and shift when blue etc) it would be the Z-axis [blue].
The VCB reports the dynamic vector - e.g. [0.0,0.0,1.0] == Z-axis.
The selected polyline's [or face's] edges are swept around in an arc: 
using the picked center, about the selected axis-of-rotation.
The edges are divided and 'faced'.
The faces will be triangulated if necessary.
If a Face is the source then the geometry has 'end-faces' - 
unless the swept arc >=360 degrees.
If a 'looped' polyline curve is the source there are no 'end-faces'.
You are asked if you want to remove coplanar faces, Yes/No...
Faces are auto-oriented.
You are finally asked if you want to reverse the faces, Yes/No...
The faced geometry is made inside a group.
The original face/edges remain unchanged.
The result is a 'true' 'Followme' swept shape...

Donations:
        Are welcome [by PayPal], please use 'TIGdonations.htm' in the 
        ../Plugins/TIGtools/ folder.
        
Version:
20090915 1.0 First issue
20090916 1.1 Pick cursor/inferencing improved, choose 'axis' added.
20090917 1.2 'Round' error fixed. Operation optimized for v7 users.
20090918 1.3 Soft/Smooth Edges option added.
20090919 1.4 Inference-locking glitch fixed on repeated lathes.
20090920 1.5 Smooth bug with Reverse Faces fixed.
20090924 1.6 Removal of Coplanar Edges now Optional. Superfluous face removal improved.
20100114 2.0 Rename extrudeEdgesByLathe, Final Explode Group and Toolbar added.
             DeBabelized and 'Extrusion Tools' Toolbar added.
20100120 2.1 ES lingvo file updated by Defisto.
20100121 2.2 Typo preventing Plugin Menu item working corrected.
20100121 2.3 FR lingvo file updated by Pilou.
20100123 2.4 FR lingvo file updated by Pilou.
20100206 2.5 Resume VCB text improved.
20100216 2.6 All extrusion-tools now in one in Plugins sub-menu.
20100222 2.7 Tooltips etc now deBabelized properly.
20100609 2.8 Glitch fixed on completion stages.
=end
###
require 'sketchup.rb'
require 'deBabelizer.rb'
###
class ExtrudeEdgesByLathe

def db(string)
  dir=File.dirname(__FILE__)+"/TIGtools"
  toolname="extrudeEdgesByLathe" 
  locale=Sketchup.get_locale.upcase
  path=dir+"/"+toolname+locale+".lingvo"
  if not File.exist?(path)
    return string
  else
    deBabelizer(string,path)
  end 
end#def

def ExtrudeEdgesByLathe::db(string)
  dir=File.dirname(__FILE__)+"/TIGtools"
  toolname="extrudeEdgesByLathe" 
  locale=Sketchup.get_locale.upcase
  path=dir+"/"+toolname+locale+".lingvo"
  if not File.exist?(path)
    return string
  else
    deBabelizer(string,path)
  end 
end#def

def initialize
  @ip=nil
  @ip1=nil
  @ip2=nil
  @state=nil
  @center=nil
  @axis=nil
  @curve=nil
end

def activate
  model=Sketchup.active_model
  @state=0
  @curve=model.selection[0]
  @face=nil
  if not @curve
    UI.messagebox(db("Extrude Edges by Lathe: Select a Polyline OR a Face..."))
    self.deactivate(model.active_view)
    return nil
  end#if
  if not @curve.class==Sketchup::Edge or not @curve.curve
    ###check to see if it's a Fsce
    @face=nil;model.selection.each{|e|
      if e.class==Sketchup::Face
        @face=e
        break
      end#if
    }
    if not @face
      UI.messagebox(db("Extrude Edges by Lathe: Select a Polyline OR a Face..."))
      self.deactivate(model.active_view)
      return nil
    end#if
  end#if
  @msg=(db("Extrude Edges by Lathe: Settings..."))
  Sketchup::set_status_text(@msg)
  Sketchup.set_status_text("",SB_VCB_VALUE)                
  Sketchup.set_status_text("", SB_VCB_LABEL)
  @angle=model.get_attribute("ExtrudeEdgesByLathe","angle",nil)
  @segs=model.get_attribute("ExtrudeEdgesByLathe","segs",nil)
  @soft=model.get_attribute("ExtrudeEdgesByLathe","soft",nil)
  @angle=45.0 if not @angle
  @segs=9 if not @segs
  @soft=(db("No")) if not @soft
  an=(db("Arc's Swept Angle: "))
  as=(db("Arc's Segments: "))
  ss=(db("Soften/Smooth ? "))
  prompts=[an,as,ss]
  values=[@angle,@segs,@soft]
  nys=(db("No|Yes|Yes+Smooth"))
  popups=["","",nys]
  title=(db("Extrude Edges by Lathe Settings"))
  results=UI.inputbox(prompts,values,popups,title)
  if not results
    self.deactivate(model.active_view) 
    return nil
  end#if
  @angle=results[0].to_f
  @segs=results[1].to_i
  @soft=results[2].to_s
  if @angle==0
    @angle=45.0
    UI.messagebox((db("Extrude Edges by Lathe: Arc's Swept Angle cannot = 0."))+"\n"+(db("Try again...")))
    self.activate
  end#if
  if @segs<1
    @segs=9
    UI.messagebox((db("Extrude Edges by Lathe: Arc's Segments must be >= 1"))+"\n"+(db("Try again...")))
    self.activate
  end#if
  @state=1
  @ip=nil
  @ip1=nil
  @ip2=nil
  @ip=Sketchup::InputPoint.new
  @ip1=Sketchup::InputPoint.new
  @ip2=Sketchup::InputPoint.new
  @msg=(db("Extrude Edges by Lathe:  Pick the Arc's Center-Point..."))
  Sketchup::set_status_text(@msg)
  Sketchup.set_status_text((db("Lathe Center")), SB_VCB_LABEL)
  Sketchup.set_status_text("",SB_VCB_VALUE)
  UI.beep
end#def

def deactivate(view)
  view.invalidate if view
  @ip=nil 
  @ip1=nil
  @ip2=nil
  @center=nil
  @axis=nil
  @curve=nil
  @state=nil
  Sketchup::set_status_text("")
  Sketchup::set_status_text("",SB_VCB_LABEL)
  Sketchup::set_status_text("",SB_VCB_VALUE)
  Sketchup.send_action("selectSelectionTool:")
  return nil
end#def

def onCancel(flag,view)
  self.deactivate(view)
  Sketchup.send_action("selectSelectionTool:")
  return nil
end#def

def resume(view)
  Sketchup::set_status_text(@msg)
  view.invalidate
end

def onMouseMove(flags,x,y,view)
  @ip.pick(view,x,y)
  view.lock_inference
  view.tooltip=(db("Lathe Center Point: "))+@ip.tooltip
  case @state
   when 1
    @msg=(db("Extrude Edges by Lathe:  Pick the Arc's Center-Point..."))
    Sketchup::set_status_text(@msg)
    Sketchup.set_status_text((db("Lathe Center")),SB_VCB_LABEL)
    if @ip.valid? and @ip!=@ip1
      @ip1.copy!(@ip)
      view.invalidate
      xyz=@ip1.position
      xyztxt="["+xyz.x.roundup(3).to_s+","+xyz.y.roundup(3).to_s+","+xyz.z.roundup(3).to_s+"]"
      Sketchup.set_status_text(xyztxt,SB_VCB_VALUE)
    end#if
   when 2
    @msg=(db("Extrude Edges by Lathe:  Pick a Second Point to Set the Axis of Rotation..."))
    Sketchup::set_status_text(@msg)
    Sketchup.set_status_text((db("Lathe Axis")),SB_VCB_LABEL)
    @ip2.pick(view,x,y,@ip1)
    if @ip2.valid? and @ip2!=@ip1
      view.invalidate
      xyz=(@ip2.position-@ip1.position).normalize
      xyztxt="["+xyz.x.roundup(3).to_s+","+xyz.y.roundup(3).to_s+","+xyz.z.roundup(3).to_s+"]"
      Sketchup.set_status_text(xyztxt,SB_VCB_VALUE)
    end#if
   when 3
     view.invalidate 
  end#case
end#def

def draw(view)
  if @ip1 and @ip1.valid? and @ip1.display? and @state<=2
    view.tooltip=(db("Lathe Center Point: "))+@ip1.tooltip
    p1=@ip1.position
    view.draw_points([p1],16,1,"orange")
    @ip1.draw(view)
    if @ip2 and @ip2.valid? and @ip2!=@ip1 and @state==2
      view.tooltip=(db("Lathe Axis Point: "))+@ip2.tooltip
      @ip2.draw(view)if @ip2.display?
      begin
        p2=@ip2.position
        view.set_color_from_line(@ip1,@ip2)
        view.draw_polyline([p1,p2])
        view.draw_points([p2],16,1,"orange")
      rescue
        ###
      end
    end#if
  end#if
end#def draw

def onKeyDown(key,repeat,flags,view)
    if key==CONSTRAIN_MODIFIER_KEY and repeat==1
        @shift_down_time=Time.now
        if view.inference_locked?
            view.lock_inference
        elsif @state==1 and @ip1.valid?
            view.lock_inference @ip1
        elsif @state==2 and @ip2.valid?
            view.lock_inference @ip2, @ip1
        end
    end
end#def

def onKeyUp(key, repeat, flags, view)
    if key==CONSTRAIN_MODIFIER_KEY and view.inference_locked? and (Time.now - @shift_down_time) > 0.5
        view.lock_inference
    end
    view.invalidate
end#def

def onLButtonDown(flags,x,y,view)
  if @state==1 ###center
    @ip1.pick view,x,y
    if @ip1.valid?
      @center=@ip1.position
      @ip.clear
      @state=2
    end#if
  else ###axis
    @ip2.pick view,x,y
    if @ip2.valid? and @ip2!=@ip1
      @axis=@ip2.position
      view.invalidate
      @state=3
      self.create_geometry()
    end#if
  end#if
end#def

def onLButtonUp(flags,x,y,view)
  view.invalidate if view
end

def create_geometry()
  Sketchup.set_status_text("",SB_VCB_VALUE)
  Sketchup.set_status_text("",SB_VCB_LABEL)
  model=Sketchup.active_model
  if Sketchup.version.split('.')[0].to_i >= 7
    model.start_operation("ExtrudeEdgesByLathe", true)
  else
    model.start_operation("ExtrudeEdgesByLathe")
  end
  entities=model.active_entities
  angle=@angle
  segs=@segs
  center=@center
  axis=@axis-@center
  points=[]
  (segs+1).times do |i|###copy curve_edge-set points and and rotate them
    pts=[]
    tr=Geom::Transformation.rotation(center,axis,(i)*(angle.degrees)/segs)
    if not @face
      @curve.curve.vertices.each{|v|
        pt=v.position
        pt.transform!(tr)
        pts<<pt
      }
    else
      @face.loops.each{|loop|
        first=nil
        loop.vertices.each{|v|
          pt=v.position
          pt.transform!(tr)
          first=pt if not first
          pts<<pt
        }
        pts<<first if first ###close loop
      }
    end#if
    points<<pts
  end#times
  group=entities.add_group()
  gents=group.entities
  ### points is array of arrays of all points
  if not @face
    len=@curve.curve.vertices.length
  else
    len=0;@face.loops.each{|loop|loop.vertices.each{|v|len+=1}}
  end#if
  vlines=[]
  num=1
  points.each{|pts|
    pts.each{|pt|
      ptnext=pts[pts.index(pt)+1]
      vline=gents.add_line(pt,ptnext)if ptnext
      vlines<<vline if vline
      @msg=((db("Extrude Edges by Lathe: Making Primary Lines "))+(num.to_s)+(db(" of "))+((segs*len).to_s))
      Sketchup::set_status_text(@msg)
      num+=1
    }
  }
  lines=[]
  num=1
  points.each{|pts|
    pts.length.times do |i|
      pt=pts[i]
      ptnext=points[points.index(pts)+1][i]if points[points.index(pts)+1]
      ptup=points[points.index(pts)+1][i+1]if points[points.index(pts)+1]
      if ptnext
        line=gents.add_line(pt,ptnext)
        lines<<line
        if ptup
          lineup=gents.add_line(pt,ptup)### = the diagonal
        end#if
        @msg=((db("Extrude Edges by Lathe: Making Secondary Lines "))+(num.to_s)+(db(" of "))+(((1+segs)*len).to_s))
        Sketchup::set_status_text(@msg)
      num+=1
      end#if
    end#do
  }
  GC.start ### ### ###
  num=1
  lines.each{|e| ### make faces
    if e and e.valid?
      @msg=((db("Extrude Edges by Lathe: Making Faces "))+(num.to_s)+(db(" of "))+((1+segs)*len).to_s)
      Sketchup::set_status_text(@msg)
      e.find_faces
      num+=1
    end#if
  }
  @msg=(db("Extrude Edges by Lathe: Tidying Faces"))
  ### check faces have 3 sides if not erase them...
  segs.times do |this|
    Sketchup::set_status_text(@msg)
    gents.each{|e|if @angle>=360
        e.erase! if e.class==Sketchup::Face and e.edges.length>3
        e.find_faces if e.class==Sketchup::Edge and e.faces.length==1
      end#if
    }
    @msg=@msg+"."
  end#times
  ### remove coplanar edges ?
  @msg=((db("Extrude Edges by Lathe: Remove Coplanar Edges ?")))
  Sketchup::set_status_text(@msg)
  cop=UI.messagebox((db("Extrude Edges by Lathe: Remove Coplanar Edges ?"))+"\n\n",MB_YESNO,"")### 6=YES 7=NO
  if cop==6
    edges=[]
    gents.each{|e|edges<<e if e.class==Sketchup::Edge}
    @msg=(db("Extrude Edges by Lathe: Removing Coplanar Edges"))
    4*segs.times do |this|
      edges.each{|e|
        e.erase! if e.valid? and e.faces.length==2 and e.faces[0].normal.dot(e.faces[1].normal) > 0.99999999999
        e.erase! if e.valid? and not e.faces[0]
      }
      Sketchup::set_status_text(@msg)
      @msg=@msg+"."
    end#times
  end#if
  ### make gable-end faces
  if @face and @angle<360
    vlines.first.find_faces if vlines.first.valid?
    vlines.last.find_faces if vlines.last.valid?
  end#if
  ### orient faces...
  faces=[];gents.each{|e|faces<<e if e.class==Sketchup::Face}
  faces.reverse! if faces[0].normal.z<0
  while faces[0]
    face=faces[0]
    face.reverse! if face.normal.z<=0
    face.orient_connected_lathed_faces
    connected=[];face.all_connected.each{|e|connected.push(e)if e.kind_of?(Sketchup::Face)}
    faces=faces-connected-[face]
  end#while
  ### soften?
  if @soft == (db("Yes"))
      gents.each{|e|
      if e.class==Sketchup::Edge
        e.soft=true
      end#if
    }
  elsif @soft == (db("Yes+Smooth"))
    gents.each{|e|
      if e.class==Sketchup::Edge
        e.soft=true
        e.smooth=true
      end#if
    }
  end#if
  model.active_view.invalidate ### ???
  ### reverse faces ?
  @msg=(db("Extrude Edges by Lathe: Reverse Faces ?"))
  Sketchup::set_status_text(@msg)
  rev=UI.messagebox((db("Extrude Edges by Lathe: Reverse Faces ?"))+"\n\n",MB_YESNO,"")### 6=YES 7=NO
  if rev==6
    faces=[];gents.each{|e|faces<<e if e.class==Sketchup::Face}
    num=1
    faces.each{|face|
      @msg=((db("Extrude Edges by Lathe: Reversing Faces "))+(num.to_s)+(db(" of "))+(faces.length.to_s))
      Sketchup::set_status_text(@msg)
      face.reverse!
      num+=1
    }
    if @soft == (db("Yes+Smooth")) ### bug fix
      gpx=entities.add_group(group)
      group.explode
      group=gpx
    end#if
  end#if
  model.active_view.invalidate ### ???
  ### tidy faceless edges...
  4.times do |this|
    group.entities.each{|e|e.erase! if e.class==Sketchup::Edge and not e.faces[0]}
  end#times
  ###
  @msg=(db("Extrude Edges by Lathe: Explode Group ?"))
  Sketchup::set_status_text(@msg)
  if UI.messagebox((db("Extrude Edges by Lathe: Explode Group ?")),MB_YESNO,"")==6 ### 6=YES 7=NO
    group.explode
  end#if
  ###
  ### set model attributes for next time used...
  model.set_attribute("ExtrudeEdgesByLathe","angle",@angle)
  model.set_attribute("ExtrudeEdgesByLathe","segs",@segs)
  model.set_attribute("ExtrudeEdgesByLathe","soft",@soft)
  ###
 model.commit_operation
  self.deactivate(model.active_view)
end#def

end#class

class Sketchup::Face
 def orient_connected_lathed_faces
    @connected_faces=[]
	self.all_connected.each{|e|
	  if e.class==Sketchup::Face
		e.edges.each{|edge|
		  if edge.faces[1]
            @connected_faces<<e
            break
          end#if
        }
	  end#if
	}
    @connected_faces=[self] + @connected_faces 
    @connected_faces.uniq!
	@awaiting_faces=@connected_faces
    @processed_faces=[self]
    @done_faces=[]
    @msg=""#(db("Extrude Edges by Lathe: Orienting Faces"))
    ###
    longstop=256;count=0
	while @awaiting_faces[0]
      @msg=@msg+"."
	  @processed_faces.each{|face|
        if not @done_faces.include?(face)
	      Sketchup::set_status_text(@msg)
		  @face=face
          face_flip
        end#if
	  }
      count+=1
      if count==longstop
       puts(db("Extrude Edges by Lathe: orient faces aborted."))
       break
     end#if
    end#while
	Sketchup::set_status_text("")
 end#def
 def face_flip
    @awaiting_faces=@awaiting_faces-[@face]
    @face.edges.each{|edge|
      rev1=edge.reversed_in?(@face)
      @common_faces=edge.faces-[@face]
      @common_faces.each{|face|
	    rev2=edge.reversed_in?(face)
        face.reverse! if @awaiting_faces.include?(face) and rev1==rev2
	    @awaiting_faces=@awaiting_faces-[face]
	    @processed_faces<<face
	  }
    }
    @done_faces<<@face
 end#def
end#class

class Float
  def roundup(x=0)
    (self*10**x).round.to_f/10**x
  end
end#def

### shortcut
def extrudeEdgesByLathe()
  Sketchup.active_model.select_tool(ExtrudeEdgesByLathe.new)
end#def

### Menu ###-----------------------------------------------
if not file_loaded?(File.basename(__FILE__))
    textstring=ExtrudeEdgesByLathe::db"Extrude Edges by Lathe"
    instructions=ExtrudeEdgesByLathe::db": Pre-Select Curve/Face, Enter Parameters, Pick Center/Axis..."
    dir=File.dirname(__FILE__)+"/TIGtools"
    toolname="extrudeEdgesByLathe"
    locale=Sketchup.get_locale.upcase
    path=dir+"/"+toolname+locale+".lingvo"
    if File.exist?(path)
      textstring=deBabelizer(textstring,path)
      instructions=deBabelizer(instructions,path)
    end#if
    cmd=UI::Command.new(textstring){extrudeEdgesByLathe()}
    if $extrusionToolsSubmenu
      $extrusionToolsSubmenu.add_item(cmd)
    else
      UI.menu("Plugins").add_item(textstring){extrudeEdgesByLathe()}
    end#if
    cmd.status_bar_text=textstring+instructions
    if $extrusionToolbar
      cmd.tooltip=textstring
      cmd.small_icon="TIGtools/extrudeEdgesByLathe16x16.png"
      cmd.large_icon="TIGtools/extrudeEdgesByLathe24x24.png"
      $extrusionToolbar.add_item(cmd)
    end#if
end
file_loaded(File.basename(__FILE__))
